/*
 * Java port of Bullet (c) 2008 Martin Dvorak <jezek2@advel.cz>
 *
 * Bullet Continuous Collision Detection and Physics Library
 * Copyright (c) 2003-2007 Erwin Coumans  http://continuousphysics.com/Bullet/
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

package javabullet.demos.opengl;

import java.nio.FloatBuffer;
import javabullet.collision.broadphase.BroadphaseNativeTypes;
import javabullet.collision.shapes.BoxShape;
import javabullet.collision.shapes.CollisionShape;
import javabullet.collision.shapes.CompoundShape;
import javabullet.collision.shapes.ConcaveShape;
import javabullet.collision.shapes.InternalTriangleIndexCallback;
import javabullet.collision.shapes.PolyhedralConvexShape;
import javabullet.collision.shapes.SphereShape;
import javabullet.collision.shapes.TriangleCallback;
import javabullet.linearmath.DebugDrawModes;
import javabullet.linearmath.Transform;
import javax.vecmath.Vector3f;
import org.lwjgl.opengl.glu.Cylinder;
import org.lwjgl.opengl.glu.Sphere;
import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.glu.GLU.*;

/**
 *
 * @author jezek2
 */
public class GLShapeDrawer {

	/*
	private static Map<CollisionShape,TriMeshKey> g_display_lists = new HashMap<CollisionShape,TriMeshKey>();
	
	private static int OGL_get_displaylist_for_shape(CollisionShape shape) {
		// JAVA NOTE: rewritten
		TriMeshKey trimesh = g_display_lists.get(shape);
		if (trimesh != null) {
			return trimesh.dlist;
		}

		return 0;
	}

	private static void OGL_displaylist_clean() {
		// JAVA NOTE: rewritten
		for (TriMeshKey trimesh : g_display_lists.values()) {
			glDeleteLists(trimesh.dlist, 1);
		}

		g_display_lists.clear();
	}
	*/

	public static void drawCoordSystem() {
		glBegin(GL_LINES);
		glColor3f(1, 0, 0);
		glVertex3f(0, 0, 0);
		glVertex3f(1, 0, 0);
		glColor3f(0, 1, 0);
		glVertex3f(0, 0, 0);
		glVertex3f(0, 1, 0);
		glColor3f(0, 0, 1);
		glVertex3f(0, 0, 0);
		glVertex3f(0, 0, 1);
		glEnd();
	}

	private static final Cylinder quadObj = new Cylinder();
	private static final Sphere sphere = new Sphere();
	
	public static void drawCylinder(float radius, float halfHeight, int upAxis) {
		glPushMatrix();
		switch (upAxis) {
			case 0:
				glRotatef(-90f, 0.0f, 1.0f, 0.0f);
				glTranslatef(0.0f, 0.0f, -halfHeight);
				break;
			case 1:
				glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
				glTranslatef(0.0f, 0.0f, -halfHeight);
				break;
			case 2:
				glTranslatef(0.0f, 0.0f, -halfHeight);
				break;
			default: {
				assert (false);
			}
		}

		// The gluCylinder subroutine draws a cylinder that is oriented along the z axis. 
		// The base of the cylinder is placed at z = 0; the top of the cylinder is placed at z=height. 
		// Like a sphere, the cylinder is subdivided around the z axis into slices and along the z axis into stacks.

		quadObj.setDrawStyle(GLU_FILL);
		quadObj.setNormals(GLU_SMOOTH);

		quadObj.draw(radius, radius, 2f * halfHeight, 15, 10);

		glPopMatrix();
	}
	
	private static int sphereDisplayList = -1;
	
	private static void drawSphere(float radius) {
		if (sphereDisplayList == -1) {
			sphereDisplayList = glGenLists(1);
			glNewList(sphereDisplayList, GL_COMPILE);
			sphere.draw(radius, 8, 8);
			glEndList();
		}
		
		glCallList(sphereDisplayList);
	}
	
	private static void glutSolidCube(float extent) {
		extent = extent * 0.5f;
		
	    glBegin(GL_QUADS);
        glNormal3f( 1f, 0f, 0f); glVertex3f(+extent,-extent,+extent); glVertex3f(+extent,-extent,-extent); glVertex3f(+extent,+extent,-extent); glVertex3f(+extent,+extent,+extent);
        glNormal3f( 0f, 1f, 0f); glVertex3f(+extent,+extent,+extent); glVertex3f(+extent,+extent,-extent); glVertex3f(-extent,+extent,-extent); glVertex3f(-extent,+extent,+extent);
        glNormal3f( 0f, 0f, 1f); glVertex3f(+extent,+extent,+extent); glVertex3f(-extent,+extent,+extent); glVertex3f(-extent,-extent,+extent); glVertex3f(+extent,-extent,+extent);
        glNormal3f(-1f, 0f, 0f); glVertex3f(-extent,-extent,+extent); glVertex3f(-extent,+extent,+extent); glVertex3f(-extent,+extent,-extent); glVertex3f(-extent,-extent,-extent);
        glNormal3f( 0f,-1f, 0f); glVertex3f(-extent,-extent,+extent); glVertex3f(-extent,-extent,-extent); glVertex3f(+extent,-extent,-extent); glVertex3f(+extent,-extent,+extent);
        glNormal3f( 0f, 0f,-1f); glVertex3f(-extent,-extent,-extent); glVertex3f(-extent,+extent,-extent); glVertex3f(+extent,+extent,-extent); glVertex3f(+extent,-extent,-extent);
		glEnd();
	}

	private static float[] childMat = new float[16];
	private static FloatBuffer wrapped;
	
	public static void drawOpenGL(float[] m, CollisionShape shape, Vector3f color, int debugMode) {
		//System.out.println("shape="+shape+" type="+BroadphaseNativeTypes.forValue(shape.getShapeType()));
		
		if (wrapped == null || wrapped.array() != m) {
			wrapped = FloatBuffer.wrap(m);
		}
		
		glPushMatrix();
		glMultMatrix(wrapped);
//		if (shape.getShapeType() == BroadphaseNativeTypes.UNIFORM_SCALING_SHAPE_PROXYTYPE.getValue())
//		{
//			const btUniformScalingShape* scalingShape = static_cast<const btUniformScalingShape*>(shape);
//			const btConvexShape* convexShape = scalingShape->getChildShape();
//			float	scalingFactor = (float)scalingShape->getUniformScalingFactor();
//			{
//				btScalar tmpScaling[4][4]={{scalingFactor,0,0,0},
//					{0,scalingFactor,0,0},
//					{0,0,scalingFactor,0},
//					{0,0,0,1}};
//
//				drawOpenGL( (btScalar*)tmpScaling,convexShape,color,debugMode);
//			}
//			glPopMatrix();
//			return;
//		}

		if (shape.getShapeType() == BroadphaseNativeTypes.COMPOUND_SHAPE_PROXYTYPE.getValue()) {
			CompoundShape compoundShape = (CompoundShape) shape;
			for (int i = compoundShape.getNumChildShapes() - 1; i >= 0; i--) {
				Transform childTrans = new Transform(compoundShape.getChildTransform(i));
				CollisionShape colShape = compoundShape.getChildShape(i);
				childTrans.getOpenGLMatrix(childMat);
				drawOpenGL(childMat, colShape, color, debugMode);
			}

		}
		else {
			//drawCoordSystem();

			//glPushMatrix();
			
			glEnable(GL_COLOR_MATERIAL);
			glColor3f(color.x, color.y, color.z);

			boolean useWireframeFallback = true;

			if ((debugMode & DebugDrawModes.DBG_DrawWireframe.getMask()) == 0) {
				switch (BroadphaseNativeTypes.forValue(shape.getShapeType())) {
					case BOX_SHAPE_PROXYTYPE: {
						BoxShape boxShape = (BoxShape) shape;
						Vector3f halfExtent = new Vector3f(boxShape.getHalfExtentsWithMargin());
						glScalef(2f * halfExtent.x, 2f * halfExtent.y, 2f * halfExtent.z);
						glutSolidCube(1f);
						useWireframeFallback = false;
						break;
					}
					case TRIANGLE_SHAPE_PROXYTYPE:
					case TETRAHEDRAL_SHAPE_PROXYTYPE: {
						//todo:	
						//					useWireframeFallback = false;
						break;
					}
					case CONVEX_HULL_SHAPE_PROXYTYPE:
						break;
					case SPHERE_SHAPE_PROXYTYPE: {
						SphereShape sphereShape = (SphereShape) shape;
						float radius = sphereShape.getMargin(); // radius doesn't include the margin, so draw with margin
						// TODO: glutSolidSphere(radius,10,10);
						//sphere.draw(radius, 8, 8);
						drawSphere(radius);
						/*
						glPointSize(10f);
						glBegin(GL_POINTS);
						glVertex3f(0f, 0f, 0f);
						glEnd();
						glPointSize(1f);
						*/
						useWireframeFallback = false;
						break;
					}
//				case CAPSULE_SHAPE_PROXYTYPE:
//				{
//					const btCapsuleShape* capsuleShape = static_cast<const btCapsuleShape*>(shape);
//					float radius = capsuleShape->getRadius();
//					float halfHeight = capsuleShape->getHalfHeight();
//					int upAxis = 1;
//
//					drawCylinder(radius,halfHeight,upAxis);
//
//
//					glPushMatrix();
//					glTranslatef(0.0, -halfHeight,0.0);
//					glutSolidSphere(radius,10,10);
//					glTranslatef(0.0, 2*halfHeight,0.0);
//					glutSolidSphere(radius,10,10);
//					glPopMatrix();
//					useWireframeFallback = false;
//					break;
//				}
					case MULTI_SPHERE_SHAPE_PROXYTYPE: {
						break;
					}
//				case CONE_SHAPE_PROXYTYPE:
//					{
//						const btConeShape* coneShape = static_cast<const btConeShape*>(shape);
//						int upIndex = coneShape->getConeUpIndex();
//						float radius = coneShape->getRadius();//+coneShape->getMargin();
//						float height = coneShape->getHeight();//+coneShape->getMargin();
//						switch (upIndex)
//						{
//						case 0:
//							glRotatef(90.0, 0.0, 1.0, 0.0);
//							break;
//						case 1:
//							glRotatef(-90.0, 1.0, 0.0, 0.0);
//							break;
//						case 2:
//							break;
//						default:
//							{
//							}
//						};
//
//						glTranslatef(0.0, 0.0, -0.5*height);
//						glutSolidCone(radius,height,10,10);
//						useWireframeFallback = false;
//						break;
//
//					}
					case CONVEX_TRIANGLEMESH_SHAPE_PROXYTYPE: {
						useWireframeFallback = false;
						break;
					}

//				case CONVEX_SHAPE_PROXYTYPE:
//				case CYLINDER_SHAPE_PROXYTYPE:
//					{
//						CylinderShape cylinder = static_cast<const btCylinderShape*>(shape);
//						int upAxis = cylinder->getUpAxis();
//
//
//						float radius = cylinder->getRadius();
//						float halfHeight = cylinder->getHalfExtentsWithMargin()[upAxis];
//
//						drawCylinder(radius,halfHeight,upAxis);
//
//						break;
//					}
					default: {
					}

				}

			}
			
			if (useWireframeFallback) {
				// for polyhedral shapes
				if (shape.isPolyhedral()) {
					PolyhedralConvexShape polyshape = (PolyhedralConvexShape) shape;

					glBegin(GL_LINES);

					Vector3f a = new Vector3f(), b = new Vector3f();
					int i;
					for (i = 0; i < polyshape.getNumEdges(); i++) {
						polyshape.getEdge(i, a, b);

						glVertex3f(a.x, a.y, a.z);
						glVertex3f(b.x, b.y, b.z);
					}
					glEnd();

//					if (debugMode==btIDebugDraw::DBG_DrawFeaturesText)
//					{
//						glRasterPos3f(0.0,  0.0,  0.0);
//						//BMF_DrawString(BMF_GetFont(BMF_kHelvetica10),polyshape->getExtraDebugInfo());
//
//						glColor3f(1.f, 1.f, 1.f);
//						for (i=0;i<polyshape->getNumVertices();i++)
//						{
//							btPoint3 vtx;
//							polyshape->getVertex(i,vtx);
//							glRasterPos3f(vtx.x(),  vtx.y(),  vtx.z());
//							char buf[12];
//							sprintf(buf," %d",i);
//							BMF_DrawString(BMF_GetFont(BMF_kHelvetica10),buf);
//						}
//
//						for (i=0;i<polyshape->getNumPlanes();i++)
//						{
//							btVector3 normal;
//							btPoint3 vtx;
//							polyshape->getPlane(normal,vtx,i);
//							btScalar d = vtx.dot(normal);
//
//							glRasterPos3f(normal.x()*d,  normal.y()*d, normal.z()*d);
//							char buf[12];
//							sprintf(buf," plane %d",i);
//							BMF_DrawString(BMF_GetFont(BMF_kHelvetica10),buf);
//
//						}
//					}


				}
			}

//		#ifdef USE_DISPLAY_LISTS
//
//		if (shape->getShapeType() == TRIANGLE_MESH_SHAPE_PROXYTYPE||shape->getShapeType() == GIMPACT_SHAPE_PROXYTYPE)
//			{
//				GLuint dlist =   OGL_get_displaylist_for_shape((btCollisionShape * )shape);
//				if (dlist)
//				{
//					glCallList(dlist);
//				}
//				else
//				{
//		#else		
			if (shape.isConcave())//>getShapeType() == TRIANGLE_MESH_SHAPE_PROXYTYPE||shape->getShapeType() == GIMPACT_SHAPE_PROXYTYPE)
			//		if (shape->getShapeType() == TRIANGLE_MESH_SHAPE_PROXYTYPE)
			{
				ConcaveShape concaveMesh = (ConcaveShape) shape;
				//btVector3 aabbMax(btScalar(1e30),btScalar(1e30),btScalar(1e30));
				//btVector3 aabbMax(100,100,100);//btScalar(1e30),btScalar(1e30),btScalar(1e30));

				//todo pass camera, for some culling
				Vector3f aabbMax = new Vector3f(1e30f, 1e30f, 1e30f);
				Vector3f aabbMin = new Vector3f(-1e30f, -1e30f, -1e30f);

				GlDrawcallback drawCallback = new GlDrawcallback();
				drawCallback.wireframe = (debugMode & DebugDrawModes.DBG_DrawWireframe.getMask()) != 0;

				concaveMesh.processAllTriangles(drawCallback, aabbMin, aabbMax);
			}
			//#endif

			//#ifdef USE_DISPLAY_LISTS
			//		}
			//	}
			//#endif

//			if (shape->getShapeType() == CONVEX_TRIANGLEMESH_SHAPE_PROXYTYPE)
//			{
//				btConvexTriangleMeshShape* convexMesh = (btConvexTriangleMeshShape*) shape;
//
//				//todo: pass camera for some culling			
//				btVector3 aabbMax(btScalar(1e30),btScalar(1e30),btScalar(1e30));
//				btVector3 aabbMin(-btScalar(1e30),-btScalar(1e30),-btScalar(1e30));
//				TriangleGlDrawcallback drawCallback;
//				convexMesh->getMeshInterface()->InternalProcessAllTriangles(&drawCallback,aabbMin,aabbMax);
//
//			}

			// TODO: error in original sources GL_DEPTH_BUFFER_BIT instead of GL_DEPTH_TEST
			glDisable(GL_DEPTH_TEST);
			glRasterPos3f(0, 0, 0);//mvtx.x(),  vtx.y(),  vtx.z());
			if ((debugMode & DebugDrawModes.DBG_DrawText.getMask()) != 0) {
				// TODO: BMF_DrawString(BMF_GetFont(BMF_kHelvetica10),shape->getName());
			}

			if ((debugMode & DebugDrawModes.DBG_DrawFeaturesText.getMask()) != 0) {
				//BMF_DrawString(BMF_GetFont(BMF_kHelvetica10),shape->getExtraDebugInfo());
			}
			glEnable(GL_DEPTH_TEST);

			//glPopMatrix();
		}
		glPopMatrix();
	}
	
	////////////////////////////////////////////////////////////////////////////
	
	private static class TriMeshKey {
		public CollisionShape shape;
		public int dlist; // OpenGL display list	
	}
	
	private static class GlDisplaylistDrawcallback implements TriangleCallback {
		private Vector3f diff1 = new Vector3f();
		private Vector3f diff2 = new Vector3f();
		private Vector3f normal = new Vector3f();
		
		public void processTriangle(Vector3f[] triangle, int partId, int triangleIndex) {
			diff1.sub(triangle[1], triangle[0]);
			diff2.sub(triangle[2], triangle[0]);
			normal.cross(diff1, diff2);

			normal.normalize();

			glBegin(GL_TRIANGLES);
			glColor3f(0, 1, 0);
			glNormal3f(normal.x, normal.y, normal.z);
			glVertex3f(triangle[0].x, triangle[0].y, triangle[0].z);

			glColor3f(0, 1, 0);
			glNormal3f(normal.x, normal.y, normal.z);
			glVertex3f(triangle[1].x, triangle[1].y, triangle[1].z);

			glColor3f(0, 1, 0);
			glNormal3f(normal.x, normal.y, normal.z);
			glVertex3f(triangle[2].x, triangle[2].y, triangle[2].z);
			glEnd();

			/*glBegin(GL_LINES);
			glColor3f(1, 1, 0);
			glNormal3d(normal.getX(),normal.getY(),normal.getZ());
			glVertex3d(triangle[0].getX(), triangle[0].getY(), triangle[0].getZ());
			glNormal3d(normal.getX(),normal.getY(),normal.getZ());
			glVertex3d(triangle[1].getX(), triangle[1].getY(), triangle[1].getZ());
			glColor3f(1, 1, 0);
			glNormal3d(normal.getX(),normal.getY(),normal.getZ());
			glVertex3d(triangle[2].getX(), triangle[2].getY(), triangle[2].getZ());
			glNormal3d(normal.getX(),normal.getY(),normal.getZ());
			glVertex3d(triangle[1].getX(), triangle[1].getY(), triangle[1].getZ());
			glColor3f(1, 1, 0);
			glNormal3d(normal.getX(),normal.getY(),normal.getZ());
			glVertex3d(triangle[2].getX(), triangle[2].getY(), triangle[2].getZ());
			glNormal3d(normal.getX(),normal.getY(),normal.getZ());
			glVertex3d(triangle[0].getX(), triangle[0].getY(), triangle[0].getZ());
			glEnd();*/
		}
	}
	
	private static class GlDrawcallback implements TriangleCallback {
		public boolean wireframe = false;
		
		public void processTriangle(Vector3f[] triangle, int partId, int triangleIndex) {
			if (wireframe) {
				glBegin(GL_LINES);
				glColor3f(1, 0, 0);
				glVertex3f(triangle[0].x, triangle[0].y, triangle[0].z);
				glVertex3f(triangle[1].x, triangle[1].y, triangle[1].z);
				glColor3f(0, 1, 0);
				glVertex3f(triangle[2].x, triangle[2].y, triangle[2].z);
				glVertex3f(triangle[1].x, triangle[1].y, triangle[1].z);
				glColor3f(0, 0, 1);
				glVertex3f(triangle[2].x, triangle[2].y, triangle[2].z);
				glVertex3f(triangle[0].x, triangle[0].y, triangle[0].z);
				glEnd();
			}
			else {
				glBegin(GL_TRIANGLES);
				glColor3f(1, 0, 0);
				glVertex3f(triangle[0].x, triangle[0].y, triangle[0].z);
				glColor3f(0, 1, 0);
				glVertex3f(triangle[1].x, triangle[1].y, triangle[1].z);
				glColor3f(0, 0, 1);
				glVertex3f(triangle[2].x, triangle[2].y, triangle[2].z);
				glEnd();
			}
		}
	}
	
	private static class TriangleGlDrawcallback implements InternalTriangleIndexCallback {
		public void internalProcessTriangleIndex(Vector3f[] triangle, int partId, int triangleIndex) {
			glBegin(GL_TRIANGLES);//LINES);
			glColor3f(1, 0, 0);
			glVertex3f(triangle[0].x, triangle[0].y, triangle[0].z);
			glVertex3f(triangle[1].x, triangle[1].y, triangle[1].z);
			glColor3f(0, 1, 0);
			glVertex3f(triangle[2].x, triangle[2].y, triangle[2].z);
			glVertex3f(triangle[1].x, triangle[1].y, triangle[1].z);
			glColor3f(0, 0, 1);
			glVertex3f(triangle[2].x, triangle[2].y, triangle[2].z);
			glVertex3f(triangle[0].x, triangle[0].y, triangle[0].z);
			glEnd();
		}
	}
	
}
